


#define EPS (1e-20)  // numerical zero
#define MIN(a,b)  ( (a>b) ? b : a )
#define MAX(a,b)  ( (a>b) ? a : b )



// DECLARATIONS:
__device__ double solve_quadprog( double *G, double *g0,
		       double *CE, double *ce0,  
		       double* CI, double* ci0, 
		       double* x,
		       int n, int p, int m );
__device__ void compute_d(double *d, double *J, const double *np,int n);
__device__ void update_z(double *z, double *J, const double *d, int iq,int n);
__device__ void update_r(double *R, double *r, const double *d, int iq,int n);
__device__ short int add_constraint(double *R, double *J, double *d, int& iq, double& R_norm,int n);
__device__ void delete_constraint(double *R, double *J, int *A, double *u, int n, int p, int& iq, int l);
__device__ double distance(double a, double b);
__device__ double scalar_product(const double *x, const double *y,int n);
__device__ void cholesky_decomposition(double *A,int n);
__device__ void cholesky_solve(double *L, double *x, const double *b,int n);
__device__ void forward_elimination(double *L, double *y, const double *b,int n);
__device__ void backward_elimination(double *U, double *x, const double *y,int n);
__device__ double distance_seg2seg( double *A1, double *A2, double *B1, double *B2 );
__device__ void print_matrix(const char* name, double *A, int n, int m);



__global__ void GLOBAL_CONST_CUDAKernel( double *DIST_CONST,
					 double *path_ALPHAs,double *path_BETAs,
					 int *FOR_LOOP_LIST,
					 int npairs)
{


  int ii = threadIdx.x + blockIdx.x * blockDim.x;  //  global thread index == pait index

  if( ii>=0 && ii<npairs ){

    int seg1 = FOR_LOOP_LIST[ 2*ii + 0 ] - 1;
    int seg2 = FOR_LOOP_LIST[ 2*ii + 1 ] - 1;

    double A1[3], A2[3], A3[3], B1[3], B2[3], B3[3], DIST;

    A1[0] = path_ALPHAs[3*seg1 + 0];  B1[0] = path_BETAs[3*seg1 + 0];
    A1[1] = path_ALPHAs[3*seg1 + 1];  B1[1] = path_BETAs[3*seg1 + 1];
    A1[2] = path_ALPHAs[3*seg1 + 2];  B1[2] = path_BETAs[3*seg1 + 2];    

    A2[0] = path_ALPHAs[3*seg2 + 0];  B2[0] = path_BETAs[3*seg2 + 0];
    A2[1] = path_ALPHAs[3*seg2 + 1];  B2[1] = path_BETAs[3*seg2 + 1];
    A2[2] = path_ALPHAs[3*seg2 + 2];  B2[2] = path_BETAs[3*seg2 + 2];    

    A3[0] = A1[0] - A2[0];  B3[0] = B1[0] - B2[0];
    A3[1] = A1[1] - A2[1];  B3[1] = B1[1] - B2[1];
    A3[2] = A1[2] - A2[2];  B3[2] = B1[2] - B2[2];

    if( scalar_product(A3,A3,3)<1e-10  && scalar_product(B3,B3,3)<1e-10 ){
      
      DIST = 0.0;
      
    }else{
      
      DIST = distance_seg2seg( A1, A2, B1, B2 );
      
    }

    DIST_CONST[ ii ] = DIST;

  }  // if


}







__device__ double distance_seg2seg( double *A1, double *A2, double *B1, double *B2 )
{

  double B1MinusB2[3] = { B1[0] - B2[0] , B1[1] - B2[1] , B1[2] - B2[2]  };

  /*
  print_matrix("A1",A1,3,1);
  print_matrix("B1",B1,3,1);
  print_matrix("A2",A2,3,1);
  print_matrix("B2",B2,3,1);
  */

  double AA = scalar_product( A1,A1,3 ); 
  double BB = scalar_product( A2,A2,3 );
  double CC = -2 * scalar_product( A1,A2,3 );
  double DD = 2 * scalar_product( A1,B1MinusB2,3 );
  double EE = -2 * scalar_product( A2,B1MinusB2,3 );
  double FF = scalar_product( B1MinusB2,B1MinusB2,3 );
  /*
  printf( "AA=%f  BB=%f  CC=%f  DD=%f  EE=%f  FF=%f\n",
	  AA,BB,CC,DD,EE,FF);
  */
  
  double H[4] = { 2*AA , CC , CC , 2*BB };  
  double LINTERM[2] = { DD , EE };
  
  double CI[8] = { 1.0 , 0.0 , 0.0 , 1.0 , -1.0 , 0.0 , 0.0 , -1.0 };
  double CI0[4] = { 0.0 , 0.0 , 1.0 , 1.0 };

  double *CE = 0, *CE0 = 0;  // NULL as p = 0

  double T[2] = { 0.5 , 0.5 };

  int n = 2;
  int p = 0;
  int m = 4;
  /*
  print_matrix("H",H,n,n);
  print_matrix("LINTERM",LINTERM,2,1);
  print_matrix("CE",CE,n,p);
  print_matrix("CE0",CE0,p,1);
  print_matrix("CI",CI,n,m);
  print_matrix("CI0",CI0,m,1);
  */
  solve_quadprog( H, LINTERM, CE, CE0, CI, CI0, T, n, p, m );

  // print_matrix("TSOL",T,n,1);  

  double XX1[3] = { T[0]*A1[0] + B1[0] , T[0]*A1[1] + B1[1] , T[0]*A1[2] + B1[2] };
  double XX2[3] = { T[1]*A2[0] + B2[0] , T[1]*A2[1] + B2[1] , T[1]*A2[2] + B2[2] };
  double XX3[3] = { XX1[0] - XX2[0] , XX1[1] - XX2[1] , XX1[2] - XX2[2] };

  double DIST = sqrt( scalar_product( XX3,XX3,3 ) );

  return DIST;

}









// The Solving function, implementing the Goldfarb-Idnani method
__device__ double solve_quadprog( double *G, double *g0,
				  double *CE, double *ce0,  
				  double* CI, double* ci0, 
				  double* x,
				  int n, int p, int m )
{
    
  int i, j, k, l;
  int ip; // this is the index of the constraint to be added to the active set
    
  double *R = (double*)malloc((n*n)*sizeof(double));
  double *J = (double*)malloc((n*n)*sizeof(double));
    
  double *s = (double*) malloc( (m + p)*sizeof(double));
  double *z = (double*) malloc(n*sizeof(double));
  double* r = (double*) malloc((m + p)*sizeof(double));
  double* d = (double*) malloc(n*sizeof(double));
  double* np = (double*) malloc(n*sizeof(double));
  double* u = (double*) malloc((m + p)*sizeof(double));
  double* x_old = (double*) malloc(n*sizeof(double));
  double* u_old = (double*) malloc((m + p)*sizeof(double));
    
  int *A = (int*)malloc((m + p)*sizeof(int));
  int *A_old = (int*)malloc((m + p)*sizeof(int));
  int *iai = (int*)malloc((m + p)*sizeof(int));
    
  double f_value, psi, c1, c2, sum, ss, R_norm;
  double inf;
  inf = 1.0E300;
  double t, t1, t2; /* t is the step lenght, which is the minimum of the partial step length t1 
		     * and the full step length t2 */
  int iq, iter = 0;
  short int* iaexcl = (short int*)malloc((m + p)*sizeof(short int));

  /* p is the number of equality constraints */
  /* m is the number of inequality constraints */
  int q = 0;  /* size of the active set A (containing the indices of the active constraints) */

  /*
   * Preprocessing phase
   */

  /* compute the trace of the original matrix G */
  c1 = 0.0;
  for (i = 0; i < n; i++)
    {
      c1 += G[i + i*n];
    }
  /* decompose the matrix G in the form L^T L */
  cholesky_decomposition(G,n);

  /* initialize the matrix R */
  for (i = 0; i < n; i++)
    {
      d[i] = 0.0;
      for (j = 0; j < n; j++)
	R[i+j*n] = 0.0;
    }
  R_norm = 1.0; /* this variable will hold the norm of the matrix R */

  /* compute the inverse of the factorized matrix G^-1, this is the initial value for H */
  c2 = 0.0;
  for (i = 0; i < n; i++) 
    {
      d[i] = 1.0;
      forward_elimination(G, z, d, n);
      for (j = 0; j < n; j++)
	J[i+j*n] = z[j];
      c2 += z[i];
      d[i] = 0.0;
    }

  /* c1 * c2 is an estimate for cond(G) */

  /* 
   * Find the unconstrained minimizer of the quadratic form 0.5 * x G x + g0 x 
   * this is a feasible point in the dual space
   * x = G^-1 * g0
   */
  cholesky_solve(G, x, g0, n);
  for (i = 0; i < n; i++)
    x[i] = -x[i];
    
  /* and compute the current solution value */ 
  f_value = 0.5 * scalar_product(g0, x, n);

  /* Add equality constraints to the working set A */
  iq = 0;
  for (i = 0; i < p; i++)
    {
      for (j = 0; j < n; j++)
	np[j] = CE[j+i*n];
        
      compute_d(d, J, np, n);
      update_z(z, J, d, iq, n);
      update_r(R, r, d, iq, n);
        
      /* compute full step length t2: i.e., the minimum step in primal space s.t. the contraint 
       * becomes feasible */
      t2 = 0.0;
      if (fabs(scalar_product(z, z, n)) > EPS) // i.e. z != 0
	t2 = (-scalar_product(np, x, n) - ce0[i]) / scalar_product(z, np, n);
        
      /* set x = x + t2 * z */
      for (k = 0; k < n; k++)
	x[k] += t2 * z[k];
        
      /* set u = u+ */
      u[iq] = t2;
      for (k = 0; k < iq; k++)
	u[k] -= t2 * r[k];
        
      /* compute the new solution value */
      f_value += 0.5 * (t2 * t2) * scalar_product(z, np, n);
      A[i] = -i - 1;
        
      if (!add_constraint(R, J, d, iq, R_norm, n))
        {  
	  // Equality constraints are linearly dependent
	  // printf("Constraints are linearly dependent");
	  goto EXIT;
        }
    }
    
  /* set iai = K \ A */
  for (i = 0; i < m; i++)
    iai[i] = i;
    
 l1:iter++;

  /* step 1: choose a violated constraint */
  for (i = p; i < iq; i++)        
    {
      ip = A[i];
      iai[ip] = -1;
    }
    
  /* compute s[x] = ci^T * x + ci0 for all elements of K \ A */
  ss = 0.0;
  psi = 0.0; /* this value will contain the sum of all infeasibilities */
  ip = 0; /* ip will be the index of the chosen violated constraint */
  for (i = 0; i < m; i++)
    {
      iaexcl[i] = true;
      sum = 0.0;
      for (j = 0; j < n; j++)
	sum += CI[j+i*n] * x[j];
      sum += ci0[i];
      s[i] = sum;
      psi += MIN(0.0, sum);
    }
    
    
  if (fabs(psi) <= m * EPS * c1 * c2* 100.0)
    {
      /* numerically there are not infeasibilities anymore */
      q = iq;        
      goto EXIT;
    }

  /* save old values for u and A */
  for (i = 0; i < iq; i++)
    {
      u_old[i] = u[i];
      A_old[i] = A[i];
    }
  /* and for x */
  for (i = 0; i < n; i++)
    x_old[i] = x[i];
    
    
 l2: /* Step 2: check for feasibility and determine a new S-pair */
        
        
  for (i = 0; i < m; i++)
    {
      if (s[i] < ss && iai[i] != -1 && iaexcl[i])
        {
	  ss = s[i];
	  ip = i;
        }
    }
  if (ss >= 0.0)
    {
      q = iq;        
      goto EXIT;
    }
    
  /* set np = n[ip] */
  for (i = 0; i < n; i++)
    np[i] = CI[i+ip*n];
    
  /* set u = [u 0]^T */
  u[iq] = 0.0;
    
  /* add ip to the active set A */
  A[iq] = ip;
    
 l2a:/* Step 2a: determine step direction */
        
  /* compute z = H np: the step direction in the primal space (through J, see the paper) */
  compute_d(d, J, np, n);
  update_z(z, J, d, iq, n);
    
  /* compute N* np (if q > 0): the negative of the step direction in the dual space */
  update_r(R, r, d, iq, n);
    

  /* Step 2b: compute step length */
  l = 0;
    
    
  /* Compute t1: partial step length (maximum step in dual space without violating dual feasibility */
  t1 = inf; /* +inf */
    
  /* find the index l s.t. it reaches the minimum of u+[x] / r */
  for (k = p; k < iq; k++)
    {
      if (r[k] > 0.0)
        {
	  if (u[k] / r[k] < t1)
            {
	      t1 = u[k] / r[k];
	      l = A[k];
            }
        }    
    }
    
    
  /* Compute t2: full step length (minimum step in primal space such that the constraint ip becomes feasible */
  if (fabs(scalar_product(z, z, n))  > EPS) // i.e. z != 0
    t2 = -s[ip] / scalar_product(z, np, n);
  else
    t2 = inf; /* +inf */

  /* the step is chosen as the minimum of t1 and t2 */
  t = MIN(t1, t2);

  /* Step 2c: determine new S-pair and take step: */

  /* case (i): no step in primal or dual space */
  if (t >= inf)
    {
      /* QPP is infeasible */
      // FIXME: unbounded to raise
      q = iq;
      f_value = inf;
      goto EXIT;
    }
  /* case (ii): step in dual space */
  if (t2 >= inf)
    {
      /* set u = u +  t * [-r 1] and drop constraint l from the active set A */
      for (k = 0; k < iq; k++)
	u[k] -= t * r[k];
        
      u[iq] += t;
      iai[l] = l;
      delete_constraint(R, J, A, u, n, p, iq, l);
      goto l2a;
    }

  /* case (iii): step in primal and dual space */

  /* set x = x + t * z */
  for (k = 0; k < n; k++)
    x[k] += t * z[k];

  /* update the solution value */
  f_value += t * scalar_product(z, np, n) * (0.5 * t + u[iq]);
    
  /* u = u + t * [-r 1] */
  for (k = 0; k < iq; k++)
    u[k] -= t * r[k];
  u[iq] += t;

  if (fabs(t - t2) < EPS)
    {
      /* full step has taken */
      /* add constraint ip to the active set*/
      if (!add_constraint(R, J, d, iq, R_norm, n))
        {
	  iaexcl[ip] = false;
	  delete_constraint(R, J, A, u, n, p, iq, ip);

	  for (i = 0; i < m; i++)
	    iai[i] = i;
	  for (i = p; i < iq; i++)
            {
	      A[i] = A_old[i];                                
	      u[i] = u_old[i];
	      iai[A[i]] = -1;
            }
	  for (i = 0; i < n; i++)
	    x[i] = x_old[i];
	  goto l2; /* go to step 2 */
        }    
      else
	iai[ip] = -1;

      goto l1;
    }

  /* a patial step has taken */
  /* drop constraint l */
  iai[l] = l;
  delete_constraint(R, J, A, u, n, p, iq, l);

  /* update s[ip] = CI * x + ci0 */
  sum = 0.0;
  for (k = 0; k < n; k++)
    sum += CI[k+ip*n] * x[k];
    
  s[ip] = sum + ci0[ip];


  goto l2a;
    
 EXIT:

  // cleanup
  free(R);
  free(J);
    
  free(s);
  free(z);
  free(r);
  free(d);
  free(np);
  free(u);
  free(x_old);
  free(u_old);
    
  free(A);
  free(A_old);
  free(iai);
    
  free(iaexcl);

  return f_value;

}



__device__ void compute_d(double *d, double *J, const double *np,int n)
{
  int i, j;
  double sum;

  /* compute d = H^T * np */
  for (i = 0; i < n; i++)
    {
      sum = 0.0;
      for (j = 0; j < n; j++)
	sum += J[j+i*n] * np[j];
      d[i] = sum;
    }
}


__device__ void update_z(double *z, double *J, const double *d, int iq,int n)
{
  int i, j;

  /* setting of z = H * d */
  for (i = 0; i < n; i++)
    {
      z[i] = 0.0;
      for (j = iq; j < n; j++)
	z[i] += J[i+j*n] * d[j];
    }
}



__device__ void update_r(double *R, double *r, const double *d, int iq,int n)
{
  int i, j;/*, n = d.size();*/
  double sum;

  /* setting of r = R^-1 d */
  for (i = iq - 1; i >= 0; i--)
    {
      sum = 0.0;
      for (j = i + 1; j < iq; j++)
	sum += R[i+j*n] * r[j];
        
      r[i] = (d[i] - sum) / R[i+i*n];
    }
}



__device__ short int add_constraint(double *R, double *J, double *d, int& iq, double& R_norm,int n)
{
  int i, j, k;
  double cc, ss, h, t1, t2, xny;

  /* we have to find the Givens rotation which will reduce the element
    d[j] to zero.
    if it is already zero we don't have to do anything, except of
    decreasing j */  
  for (j = n - 1; j >= iq + 1; j--)
    {
      /* The Givens rotation is done with the matrix (cc cs, cs -cc).
       * If cc is one, then element (j) of d is zero compared with element
       * (j - 1). Hence we don't have to do anything. 
       * If cc is zero, then we just have to switch column (j) and column (j - 1) 
       * of J. Since we only switch columns in J, we have to be careful how we
       * update d depending on the sign of gs.
       * Otherwise we have to apply the Givens rotation to these columns.
       * The i - 1 element of d has to be updated to h. */
      cc = d[j - 1];
      ss = d[j];
      h = distance(cc, ss);
      if (fabs(h) < EPS) // h == 0 
	continue;
        
      d[j] = 0.0;
      ss = ss / h;
      cc = cc / h;
      if (cc < 0.0)
        {
	  cc = -cc;
	  ss = -ss;
	  d[j - 1] = -h;
        }
      else
	d[j - 1] = h;
        
      xny = ss / (1.0 + cc);
      for (k = 0; k < n; k++)
        {
	  t1 = J[k+(j - 1)*n];
	  t2 = J[k+j*n];
	  J[k+(j - 1)*n] = t1 * cc + t2 * ss;
	  J[k+j*n] = xny * ( t1 + J[ k+(j-1)*n ] ) - t2;
        }
    }
  /* update the number of constraints added*/
  iq++;
  /* To update R we have to put the iq components of the d vector
   * into column iq - 1 of R
   */
  for (i = 0; i < iq; i++)
    R[i+(iq - 1)*n] = d[i];
    
  if (fabs(d[iq - 1]) <= EPS * R_norm) 
    {        
      // problem degenerate
      return false;
    }
  R_norm = MAX(R_norm, fabs(d[iq - 1]));
  return true;
}



__device__ void delete_constraint(double *R, double *J, int *A, double *u, int n, int p, int& iq, int l)
{

  int i, j, k, qq = -1; // just to prevent warnings from smart compilers
  double cc, ss, h, xny, t1, t2;

  /* Find the index qq for active constraint l to be removed */
  for (i = p; i < iq; i++)
    if (A[i] == l)
      {
	qq = i;
	break;
      }
    
  /* remove the constraint from the active set and the duals */
  for (i = qq; i < iq - 1; i++)
    {
      A[i] = A[i + 1];
      u[i] = u[i + 1];
      for (j = 0; j < n; j++)
	R[j+n*i] = R[j+(i + 1)*n];
    }
    
  A[iq - 1] = A[iq];
  u[iq - 1] = u[iq];
  A[iq] = 0; 
  u[iq] = 0.0;
  for (j = 0; j < iq; j++)
    R[j+(iq - 1)*n] = 0.0;
    
  /* constraint has been fully removed */
  iq--;
    
  if (iq == 0)
    return;
    
  for (j = qq; j < iq; j++)
    {
      cc = R[j+j*n];
        
      ss = R[j + 1 + j*n];
      h = distance(cc, ss);
      if (fabs(h) < EPS) // h == 0
	continue;
      cc = cc / h;
      ss = ss / h;
      R[j + 1 + j*n] = 0.0;
      if (cc < 0.0)
        {
	  R[j+j*n] = -h;
	  cc = -cc;
	  ss = -ss;
        }
      else            
	R[j+j*n] = h;
        
      xny = ss / (1.0 + cc);
      for (k = j + 1; k < iq; k++)
        {
	  t1 = R[j+k*n];
	  t2 = R[j + 1 + k*n];
	  R[j+k*n] = t1 * cc + t2 * ss;
	  R[j + 1 + k*n] = xny * (t1 + R[j+k*n]) - t2;
        }
      for (k = 0; k < n; k++)
        {
	  t1 = J[k+j*n];
	  t2 = J[k + (j + 1)*n];
	  J[k + j*n] = t1 * cc + t2 * ss;
	  J[k + (j + 1)*n] = xny * (J[k + j*n] + t1) - t2;
        }
    }
}



__device__ double distance(double a, double b)
{
  double a1, b1, t;
  a1 = fabs(a);
  b1 = fabs(b);
  if (a1 > b1) 
    {
      t = (b1 / a1);
      return a1 * sqrt(1.0 + t * t);
    }
  else
    if (b1 > a1)
      {
	t = (a1 / b1);
	return b1 * sqrt(1.0 + t * t);
      }
  return a1 * sqrt(2.0);
}



__device__ double scalar_product(const double *x, const double *y,int n)
{
  int i;
  double sum;
    
  sum = 0.0;
  for (i = 0; i < n; i++)
    sum += x[i] * y[i];
    
  return sum;
}



__device__ void cholesky_decomposition(double *A,int n) 
{
  int i, j, k;
  double sum;
    
  for (i = 0; i < n; i++)
    {
      for (j = i; j < n; j++)
        {
	  sum = A[i+j*n];
	  for (k = i - 1; k >= 0; k--)
	    sum -= A[i+k*n] * A[j+k*n];
            
	  if (i == j) 
            {
	      if (sum <= 0.0)                    
                {
		  goto END;
                }
	      A[i+i*n] = sqrt(sum);
            }
	  else
	    A[j+i*n] = sum / A[i+i*n];
        }
      for (k = i + 1; k < n; k++)
	A[i+k*n] = A[k+i*n];
    } 


 END: 
  ;
}



__device__ void cholesky_solve(double *L, double *x, const double *b,int n)
{
  double *y = (double *)malloc( n*sizeof(double) );

  /* Solve L * y = b */    
  forward_elimination(L, y, b, n);
    
  /* Solve L^T * x = y */
  backward_elimination(L, x, y, n);
    
  free(y);
}





__device__ void forward_elimination(double *L, double *y, const double *b,int n)
{
  int i, j;

  y[0] = b[0] / L[0];
  for (i = 1; i < n; i++)
    {
      y[i] = b[i];
      for (j = 0; j < i; j++)
	y[i] -= L[i+j*n] * y[j];
      y[i] = y[i] / L[i+i*n];
    }
}



__device__ void backward_elimination(double *U, double *x, const double *y,int n)
{
  int i, j;

  x[n - 1] = y[n - 1] / U[n - 1 + (n - 1)*n];
  for (i = n - 2; i >= 0; i--)
    {
      x[i] = y[i];
      for (j = i + 1; j < n; j++)
	x[i] -= U[i+j*n] * x[j];
      x[i] = x[i] / U[i+i*n];
    }
}





__device__ void print_matrix(const char* name, double *A, int n, int m)
{
    
  printf("%s :\n",name);
    
  for (int i = 0; i < n; i++)
    {
      printf("  ");
      for (int j = 0; j < m; j++)
	printf("%f , ",A[i+j*n]);
      printf("\n");
    }
}
