






function [ DCP2SEG C_COMB ] = print_path_constraints( x,Dmin,Kmax,Nrej,string2print )


N = size(x,1) /3 ;
path = reshape(x,N,3);

path_PARAMS = compute_linear_params( path );

Nseg = N-1;
DCP2SEG = -ones( Nseg,Nseg );

figure; 
subplot(2,2,1); plot3( path(:,1),path(:,2),path(:,3),'-ok','LineWidth',2 );
axis image; hold on;

colors = 'rbgmcy';
Ncolors = size(colors,2);

for ii = 1 : Nseg  % loop over segments

    % the distance between N and M is the same as the distance between M
    % and N, therefore the distance matrix is symmetric -- FLAG = 1    
    FLAG = 1;
    
    [ DIST PT1 PT2 ] = distance_seg2AllSegs( path_PARAMS,ii,FLAG );

    DCP2SEG( ii,: ) = DIST;
    
    for jj = 1 : Nseg
        if jj > ii + (Nrej-1)  && DIST( jj ) < Dmin
            xxx = [ PT1(jj,1);PT2(jj,1) ];
            yyy = [ PT1(jj,2);PT2(jj,2) ];
            zzz = [ PT1(jj,3);PT2(jj,3) ];
            
            eval( sprintf( 'plot3( xxx,yyy,zzz,''-%s'' );',colors(mod(ii-1,Ncolors)+1) ) );
        end
    end
    
end


ONES = -ones( Nseg,Nseg );
DCP2SEG = full( spdiags(ONES,[0:Nrej-1],DCP2SEG) );


title( sprintf('%s',string2print) );

subplot(2,2,2);
imagesc(DCP2SEG); axis image; colorbar;
title( sprintf('SEGMENT DISTANCE MATRIX (%s)',string2print) ); 
xlabel('SEGMENT #'); ylabel('SEGMENT #');

C_min = min( DCP2SEG(:) );
C_max = max( DCP2SEG(:) );

nCmap = 256;

n1 = int32( ( 0 - C_min ) / ( C_max - C_min ) * nCmap );
n2 = int32( ( Dmin - 0 ) / ( C_max - C_min ) * nCmap );
n3 = int32( ( C_max - Dmin ) / ( C_max - C_min ) * nCmap );

while ( n1+n2+n3 ) > nCmap
    n3 = n3 - 1;
end

while ( n1+n2+n3 ) < nCmap
    n3 = n3 + 1;
end

C_gray = colormap( gray(256) );
C_hot = colormap( jet(256) );

C_COMB = [ zeros(n1,3) ; interp1( linspace(0,1,nCmap)',C_hot,linspace(0,1,n2) ) ; interp1( linspace(0,1,nCmap)',C_gray,linspace(0,1,n3) ) ];  % combine hot and gray scales

colormap( C_COMB );


% CURVATURE
dS = sqrt( sum( abs( path(2:end,:) - path(1:end-1,:) ).^2,2 ) );

T = path(2:end,:) - path(1:end-1,:);  % tangent vector
Tnorm = sqrt( sum( abs(T).^2,2 ) );
T = T ./ repmat(Tnorm,[1 3]);  % normalize (unit vector)

K = T(2:end,:) - T(1:end-1,:);
K = K ./ repmat(dS(1:end-1),[1 3]);  % K = dT/dS
Knorm = sqrt( sum( abs(K).^2,2 ) );

subplot(2,2,3);
plot( Knorm,'-ob' ); hold on;

x = [ 1 : Nseg-1 ]';
y = ones(Nseg-1,1) * Kmax;

plot(x,y,'-r','LineWidth',2);

xlabel('SEGMENT #'); title( sprintf('CURVATURE (%s)',string2print) );

legend('K','Kmax');

axis([ 1 Nseg-1  0 max([Kmax ; Knorm])*1.2 ]);












