




function [ path_opt DCP2SEG_pre DCP2SEG_post cmap_pre cmap_post ] = deform_path_LINEAR( path,n_resample,Dmin,Kmax,display,nCPUs )
% function [ path_opt DCP2SEG_pre DCP2SEG_post cmap_pre cmap_post ] = deform_path_LINEAR( path,n_resample,Dmin,Kmax,display,nCPUs )

global gpuStruct;

gpuStruct = [];

close all;

if n_resample>0
    path = resample_path( path,n_resample );
end


N = size(path,1);
Nseg = N-1;

path_PARAMS = compute_linear_params( path );

% determine # of segments to be rejected from the min distance constraints
% to avoid taking into account neighborhing segments
% av_DIST_btw_segs = mean( sqrt( sum( abs( path(2:end,:) - path(1:end-1,:) ).^2,2 ) ) );
min_DIST_btw_segs = min( sqrt( sum( abs( path(2:end,:) - path(1:end-1,:) ).^2,2 ) ) );

nSEG_rej = ceil( Dmin / min_DIST_btw_segs ) * 3;

nSEG_rej = max( nSEG_rej,2 ); % NEEDS TO BE AT LEAST 2. Indeed, the distance between segments N and N is 0
                                % between N and (N-1) is 0 and between N
                                % and (N+1) is also 0

nSEG_rej = 3;
                                
path0 = path ;  % .* (1 + 2*1e-1*(rand(size(path))-0.5) );
X0 = path0(:);
% X0 = X0 .* ( 1 + 2.0*1e-1*(rand(size(X0))-0.5) );


% FOR_LOOP_LIST is the list of seg-seg indices at which to estimate the
% constraints. It is nothing more that the upper triangular portion of a square
% matrix of size Nseg*Nseg with the diagonal and the next diagonal up
% removed (this is because the distance between seg-seg and seg-(seg+1)
% is 0...
TEMP_1 = ones( Nseg,Nseg );
TEMP_0 = zeros( Nseg,Nseg );
TEMP_1 = full( spdiags(TEMP_0,[ 0 : (nSEG_rej-1) ],TEMP_1) );
IND = find( triu( TEMP_1 ) > 0 );
[ ind1 ind2 ] = ind2sub( [Nseg Nseg],IND );

FOR_LOOP_LIST = [ ind1 ind2 IND ];


FUN = @(x)GLOBAL_FUN( x,X0 );

if nCPUs == 1  % SINGLE CPU
    
    fprintf('Optimizing path using 1 CPU ...\n');
    
    FUN_CONST = @(x)GLOBAL_CONST( x,Dmin,Kmax,FOR_LOOP_LIST );
    
elseif nCPUs > 1  % MULTIPLE CPUS (6x faster than single CPU when nCPUS=8)
    
    fprintf('Optimizing path using parallel computing (%d CPUs) ...\n',nCPUs);

    poolobj = gcp('nocreate');
    if isempty(poolobj)
        poolsize = 0;
    else
        poolsize = poolobj.NumWorkers;
    end
       
    if poolsize ~= nCPUs
        parpool('local',nCPUs);
    end
    
    FUN_CONST = @( x )GLOBAL_CONST_par( x,Dmin,Kmax,FOR_LOOP_LIST );
    
else  % GPU (72x faster than single CPU -- most of the computation time is dominated by transfer of 
      %          data bask from GPU to CPU)
    fprintf('Optimizing path using parallel computing (GPU) ...\n',nCPUs);
    
    prepare_GPU_kernels( path_PARAMS,FOR_LOOP_LIST );  % update gpuStruct
    
    FUN_CONST = @( x )GLOBAL_CONST_GPU( x,Dmin,Kmax );

end

UB = path + Dmin;
LB = path - Dmin;


DCP2SEG_pre = [];
DCP2SEG_post = [];

if display ==1
    
    fprintf('Analyzing initial path ...\n');
    
    if nCPUs == 1
        [DCP2SEG_pre cmap_pre] = analyze_initial_path( X0,Dmin,Kmax,nSEG_rej,'INITIAL PATH' );
    else
        [DCP2SEG_pre cmap_pre] = analyze_initial_path_par( X0,Dmin,Kmax,nSEG_rej,'INITIAL PATH',FOR_LOOP_LIST );
    end
    
    drawnow;
    
end


fprintf('\n\n*** OPTIMIZATION STARTS ***\n');

% OPTIONS = optimoptions( 'fmincon' , 'Algorithm','active-set', 'Display','iter-detailed' , 'GradObj','off' , 'GradConstr','off' , ...
%     'MaxFunEvals',1e6 , 'MaxIter',1e6 , 'TolCon',1e-5 , 'TolFun',1e-20 );

OPTIONS = optimoptions( 'fmincon' , 'Algorithm','interior-point', 'Display','iter-detailed' , 'GradObj','off' , 'GradConstr','off' , ...
    'MaxFunEvals',1e6 , 'MaxIter',1e6 , 'TolCon',1e-5 , 'TolFun',1e-20 );

% OPT_PROBLEM = createOptimProblem( 'fmincon' , 'nonlcon',FUN_CONST , 'objective',FUN , 'x0',X0 , 'options',OPTIONS , ...
%                 'ub',UB , 'lb',LB  );
% 
% GS = GlobalSearch( 'Display','iter' , 'TolFun',1e-4 , 'TolX',1e-4 );
% 
% [ x fval exitflag output solutions ] = run( GS,OPT_PROBLEM );
tic;
[ x fval ] = fmincon( FUN,X0,[],[],[],[],LB,UB,FUN_CONST,OPTIONS );
T2 = toc;

fprintf('*** OPTIMIZATION ENDS (%f minutes) ***\n\n\n',T2/60);

path_opt = reshape(x,N,3);

if display ==1

    figure;
    plot3(path0(:,1),path0(:,2),path0(:,3),'-or','LineWidth',2);

    hold on;
    plot3(path_opt(:,1),path_opt(:,2),path_opt(:,3),'-ob','LineWidth',2);
    
    DISP = path_opt - path0;
    quiver3( path0(:,1),path0(:,2),path0(:,3),DISP(:,1),DISP(:,2),DISP(:,3),0,'-k' );
    axis image;
    
    legend('PATH0','PATH_opt','Disp.');


    if nCPUs == 1
        [DCP2SEG_post cmap_post] = analyze_initial_path( x,Dmin,Kmax,nSEG_rej,'PATH OPTIMIZED' );
    else
        [DCP2SEG_post cmap_post] = analyze_initial_path_par( x,Dmin,Kmax,nSEG_rej,'PATH OPTIMIZED',FOR_LOOP_LIST );
    end
    
end


















    





























