


clear;
close all;

addpath( '/autofs/cluster/wald/6/guerin/MGH_DBS_PATIENT_IMAGING_DATA/DBS_RemoveLead_Intersections/CODE' );


% load uncorrected, numbered chain for bilaterial DBS
load chain_1;
chain1 = chain;

load chain_2;
chain2 = chain;

L1 = length_path(chain1);  % mm
L2 = length_path(chain2);  % mm

% multiresolution resolutions -- first pass is 5.0 mm, second pass is 3.0 mm         
n_resample1 = ceil( L1 ./ [ 5.0 3.0 ] );  % lead #1
n_resample2 = ceil( L2 ./ [ 5.0 3.0 ] );  % lead #2

Dmin = 2.0 * 1.01;  % DISTANCE CONSTRAINT IN MM, should be equal to the diameter of the lead + 1% margin
Kmax = 1/1.0;  % CURVATURE CONSTRAINT in 1/MM, should be equal to 1/radius of the lead

display = 1;  % 1: display  //  0: no display

nCPUs = -8;  % <0: GPU  //  1: One CPU  //  >1: multiple CPUs


% multi-resolution loop
for ii = 1 : size(n_resample1,2)
    
    fprintf('\n***************************************************\n');
    fprintf('*    MULTI-RESOLUTION LOOP #%d (Npoints = %d)     *\n',ii,n_resample1(ii)+n_resample2(ii));
    fprintf('***************************************************\n\n');
    
    % resample chains 1 & 2 @ required resolution
    chain10 = resample_path( chain1,n_resample1(ii) );
    chain20 = resample_path( chain2,n_resample2(ii) );

    % concatenate chains
    chain_tot = [ chain10 ; chain20 ];
    
    % optimized global chain
    chain = deform_path_LINEAR( chain_tot,-1,Dmin,Kmax,display,nCPUs );
    
    % separate chains
    chain1 = chain( 1:n_resample1(ii) , : );
    chain2 = chain( n_resample1(ii)+1 : end , : );
    
end

chain0 = chain10;
chain = chain1;
save PATH_OPT_BILATERAL_chain1.mat chain chain0;

chain0 = chain20;
chain = chain2;
save PATH_OPT_BILATERAL_chain2.mat chain chain0;









