
clear;
close all;

fname = select_file('*.mat','./','SELECT DBS PATH CHAIN');
load(fname);

pitch = input('Please enter helix PITCH: ');
radius = input( 'Please enter helix RADIUS: ' );

% chain1 = [ zeros(100,2) linspace(0,10,100)' ];
% chain2 = [ linspace(0,10,100)' zeros(100,1)  ones(100,1)*10 ];
% chain3 = [ ones(100,1)*10 zeros(100,1) linspace(10,0,100)' ];
% chain = [ chain1 ; chain2 ; chain3 ];

pitch = 1;
radius = 0.5;

A = radius;
B = pitch / (2*pi);

% interpolate lead path with adequate resolution:
[ CS IND ]= unique( [ 0; cumsum( sqrt ( sum( diff(chain,[],1 ).^2,2 ) ) ) ] );

L = CS(end);
N = ceil( L / pitch * 10 );  % we want at leats 20 points per complete turn of the helix

chain_INT = interp1( CS,chain(IND,:),linspace(0,L,N),'linear' );

chain_INT_DIFF = chain_INT(2:end,:) - chain_INT(1:end-1,:);
chain_INT_DIFF = [ chain_INT_DIFF ; chain_INT_DIFF(end,:) ];


% create helix:
[ CS IND ]= unique( [ 0; cumsum( sqrt ( sum( diff(chain_INT,[],1 ).^2,2 ) ) ) ] );

chain_HELIX = [];

for ii = 1 : N
    
    % compute angle helix based on the current pitch (pitch = 
    % curvilinear coordinate on the path @ current point)
    theta = CS(ii) / B;
    
    % tengantial vector of the path @ current point
    if ii >= 4  && ii <= (N-3)
        % T = mean( chain_INT_DIFF(ii-3:ii+3,:),1 );  % smooth tangente over several points to avoid sharp angles (causes helix artifacts)
        T = mean( chain_INT_DIFF(ii,:),1 );  % smooth tangente over several points to avoid sharp angles (causes helix artifacts)
    else
        T = chain_INT_DIFF(ii,:);
    end
    T = T / norm(T);
    
    THE = acos( T(3) );
    PHI = atan2( T(2),T(1) );
    
    Ry = [ cos(THE) 0   sin(THE) ;
            0       1   0        ;
          -sin(THE) 0   cos(THE) ];
      
    Rz = [ cos(PHI)  -sin(PHI)  0;
           sin(PHI)  cos(PHI)   0;
            0          0        1 ];
    
    X = [ A*cos(theta) ; A*sin(theta) ; 0 ];  % helix along z
    
    X_trans = Rz * Ry * X   +   chain_INT(ii,:)';
    
    chain_HELIX = [ chain_HELIX ; X_trans' ] ;
    
end


f = figure;
plot3( chain_INT(:,1),chain_INT(:,2),chain_INT(:,3),'-ob','LineWidth',3 );
hold on;
plot3( chain_HELIX(:,1),chain_HELIX(:,2),chain_HELIX(:,3),'-or','LineWidth',3 ); axis image off;


chain = chain_HELIX( 10:end-10,: );
eval( sprintf('save %s_HELIX_COND.mat chain;',fname(3:end-4)) );








